@@ -2,7 +2,7 @@
 Open Source Initiative OSI - The MIT License (MIT):Licensing
 
 The MIT License (MIT)
-Copyright (c) 2013 Ralph Caraveo (deckarep@gmail.com)
+Copyright (c) 2013 - 2022 Ralph Caraveo (deckarep@gmail.com)
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
@@ -25,34 +25,59 @@ SOFTWARE.
 
 package mapset
 
-import "testing"
+import (
+	"testing"
+)
 
-func makeSet(ints []int) Set {
-	set := NewSet()
+func makeSetInt(ints []int) Set[int] {
+	s := NewSet[int]()
 	for _, i := range ints {
-		set.Add(i)
+		s.Add(i)
 	}
-	return set
+	return s
 }
 
-func makeUnsafeSet(ints []int) Set {
-	set := NewThreadUnsafeSet()
+func makeUnsafeSetInt(ints []int) Set[int] {
+	s := NewThreadUnsafeSet[int]()
 	for _, i := range ints {
-		set.Add(i)
+		s.Add(i)
+	}
+	return s
+}
+
+func makeSetIntWithAppend(ints ...int) Set[int] {
+	s := NewSet[int]()
+	s.Append(ints...)
+	return s
+}
+
+func makeUnsafeSetIntWithAppend(ints ...int) Set[int] {
+	s := NewThreadUnsafeSet[int]()
+	s.Append(ints...)
+	return s
+}
+
+func assertEqual[T comparable](a, b Set[T], t *testing.T) {
+	if !a.Equal(b) {
+		t.Errorf("%v != %v\n", a, b)
 	}
-	return set
 }
 
 func Test_NewSet(t *testing.T) {
-	a := NewSet()
-
+	a := NewSet[int]()
 	if a.Cardinality() != 0 {
 		t.Error("NewSet should start out as an empty set")
 	}
+
+	assertEqual(NewSet([]int{}...), NewSet[int](), t)
+	assertEqual(NewSet([]int{1}...), NewSet(1), t)
+	assertEqual(NewSet([]int{1, 2}...), NewSet(1, 2), t)
+	assertEqual(NewSet([]string{"a"}...), NewSet("a"), t)
+	assertEqual(NewSet([]string{"a", "b"}...), NewSet("a", "b"), t)
 }
 
 func Test_NewUnsafeSet(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 
 	if a.Cardinality() != 0 {
 		t.Error("NewSet should start out as an empty set")
@@ -60,7 +85,7 @@ func Test_NewUnsafeSet(t *testing.T) {
 }
 
 func Test_AddSet(t *testing.T) {
-	a := makeSet([]int{1, 2, 3})
+	a := makeSetInt([]int{1, 2, 3})
 
 	if a.Cardinality() != 3 {
 		t.Error("AddSet does not have a size of 3 even though 3 items were added to a new set")
@@ -68,15 +93,31 @@ func Test_AddSet(t *testing.T) {
 }
 
 func Test_AddUnsafeSet(t *testing.T) {
-	a := makeUnsafeSet([]int{1, 2, 3})
+	a := makeUnsafeSetInt([]int{1, 2, 3})
 
 	if a.Cardinality() != 3 {
 		t.Error("AddSet does not have a size of 3 even though 3 items were added to a new set")
 	}
 }
 
+func Test_AppendSet(t *testing.T) {
+	a := makeSetIntWithAppend(1, 2, 3)
+
+	if a.Cardinality() != 3 {
+		t.Error("AppendSet does not have a size of 3 even though 3 items were added to a new set")
+	}
+}
+
+func Test_AppendUnsafeSet(t *testing.T) {
+	a := makeUnsafeSetIntWithAppend(1, 2, 3)
+
+	if a.Cardinality() != 3 {
+		t.Error("AppendSet does not have a size of 3 even though 3 items were added to a new set")
+	}
+}
+
 func Test_AddSetNoDuplicate(t *testing.T) {
-	a := makeSet([]int{7, 5, 3, 7})
+	a := makeSetInt([]int{7, 5, 3, 7})
 
 	if a.Cardinality() != 3 {
 		t.Error("AddSetNoDuplicate set should have 3 elements since 7 is a duplicate")
@@ -88,7 +129,7 @@ func Test_AddSetNoDuplicate(t *testing.T) {
 }
 
 func Test_AddUnsafeSetNoDuplicate(t *testing.T) {
-	a := makeUnsafeSet([]int{7, 5, 3, 7})
+	a := makeUnsafeSetInt([]int{7, 5, 3, 7})
 
 	if a.Cardinality() != 3 {
 		t.Error("AddSetNoDuplicate set should have 3 elements since 7 is a duplicate")
@@ -99,8 +140,32 @@ func Test_AddUnsafeSetNoDuplicate(t *testing.T) {
 	}
 }
 
+func Test_AppendSetNoDuplicate(t *testing.T) {
+	a := makeSetIntWithAppend(7, 5, 3, 7)
+
+	if a.Cardinality() != 3 {
+		t.Error("AppendSetNoDuplicate set should have 3 elements since 7 is a duplicate")
+	}
+
+	if !(a.Contains(7) && a.Contains(5) && a.Contains(3)) {
+		t.Error("AppendSetNoDuplicate set should have a 7, 5, and 3 in it.")
+	}
+}
+
+func Test_AppendUnsafeSetNoDuplicate(t *testing.T) {
+	a := makeUnsafeSetIntWithAppend(7, 5, 3, 7)
+
+	if a.Cardinality() != 3 {
+		t.Error("AppendSetNoDuplicate set should have 3 elements since 7 is a duplicate")
+	}
+
+	if !(a.Contains(7) && a.Contains(5) && a.Contains(3)) {
+		t.Error("AppendSetNoDuplicate set should have a 7, 5, and 3 in it.")
+	}
+}
+
 func Test_RemoveSet(t *testing.T) {
-	a := makeSet([]int{6, 3, 1})
+	a := makeSetInt([]int{6, 3, 1})
 
 	a.Remove(3)
 
@@ -120,8 +185,28 @@ func Test_RemoveSet(t *testing.T) {
 	}
 }
 
+func Test_RemoveAllSet(t *testing.T) {
+	a := makeSetInt([]int{6, 3, 1, 8, 9})
+
+	a.RemoveAll(3, 1)
+
+	if a.Cardinality() != 3 {
+		t.Error("RemoveAll should only have 2 items in the set")
+	}
+
+	if !a.Contains(6, 8, 9) {
+		t.Error("RemoveAll should have only items (6,8,9) in the set")
+	}
+
+	a.RemoveAll(6, 8, 9)
+
+	if a.Cardinality() != 0 {
+		t.Error("RemoveSet should be an empty set after removing 6 and 1")
+	}
+}
+
 func Test_RemoveUnsafeSet(t *testing.T) {
-	a := makeUnsafeSet([]int{6, 3, 1})
+	a := makeUnsafeSetInt([]int{6, 3, 1})
 
 	a.Remove(3)
 
@@ -141,8 +226,28 @@ func Test_RemoveUnsafeSet(t *testing.T) {
 	}
 }
 
+func Test_RemoveAllUnsafeSet(t *testing.T) {
+	a := makeUnsafeSetInt([]int{6, 3, 1, 8, 9})
+
+	a.RemoveAll(3, 1)
+
+	if a.Cardinality() != 3 {
+		t.Error("RemoveAll should only have 2 items in the set")
+	}
+
+	if !a.Contains(6, 8, 9) {
+		t.Error("RemoveAll should have only items (6,8,9) in the set")
+	}
+
+	a.RemoveAll(6, 8, 9)
+
+	if a.Cardinality() != 0 {
+		t.Error("RemoveSet should be an empty set after removing 6 and 1")
+	}
+}
+
 func Test_ContainsSet(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 
 	a.Add(71)
 
@@ -166,7 +271,7 @@ func Test_ContainsSet(t *testing.T) {
 }
 
 func Test_ContainsUnsafeSet(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 
 	a.Add(71)
 
@@ -190,7 +295,7 @@ func Test_ContainsUnsafeSet(t *testing.T) {
 }
 
 func Test_ContainsMultipleSet(t *testing.T) {
-	a := makeSet([]int{8, 6, 7, 5, 3, 0, 9})
+	a := makeSetInt([]int{8, 6, 7, 5, 3, 0, 9})
 
 	if !a.Contains(8, 6, 7, 5, 3, 0, 9) {
 		t.Error("ContainsAll should contain Jenny's phone number")
@@ -202,7 +307,7 @@ func Test_ContainsMultipleSet(t *testing.T) {
 }
 
 func Test_ContainsMultipleUnsafeSet(t *testing.T) {
-	a := makeUnsafeSet([]int{8, 6, 7, 5, 3, 0, 9})
+	a := makeUnsafeSetInt([]int{8, 6, 7, 5, 3, 0, 9})
 
 	if !a.Contains(8, 6, 7, 5, 3, 0, 9) {
 		t.Error("ContainsAll should contain Jenny's phone number")
@@ -213,8 +318,88 @@ func Test_ContainsMultipleUnsafeSet(t *testing.T) {
 	}
 }
 
+func Test_ContainsOneSet(t *testing.T) {
+	a := NewSet[int]()
+
+	a.Add(71)
+
+	if !a.ContainsOne(71) {
+		t.Error("ContainsSet should contain 71")
+	}
+
+	a.Remove(71)
+
+	if a.ContainsOne(71) {
+		t.Error("ContainsSet should not contain 71")
+	}
+
+	a.Add(13)
+	a.Add(7)
+	a.Add(1)
+
+	if !(a.ContainsOne(13) && a.ContainsOne(7) && a.ContainsOne(1)) {
+		t.Error("ContainsSet should contain 13, 7, 1")
+	}
+}
+
+func Test_ContainsOneUnsafeSet(t *testing.T) {
+	a := NewThreadUnsafeSet[int]()
+
+	a.Add(71)
+
+	if !a.ContainsOne(71) {
+		t.Error("ContainsSet should contain 71")
+	}
+
+	a.Remove(71)
+
+	if a.ContainsOne(71) {
+		t.Error("ContainsSet should not contain 71")
+	}
+
+	a.Add(13)
+	a.Add(7)
+	a.Add(1)
+
+	if !(a.ContainsOne(13) && a.ContainsOne(7) && a.ContainsOne(1)) {
+		t.Error("ContainsSet should contain 13, 7, 1")
+	}
+}
+
+func Test_ContainsAnySet(t *testing.T) {
+	a := NewSet[int]()
+
+	a.Add(71)
+
+	if !a.ContainsAny(71) {
+		t.Error("ContainsSet should contain 71")
+	}
+
+	if !a.ContainsAny(71, 10) {
+		t.Error("ContainsSet should contain 71 or 10")
+	}
+
+	a.Remove(71)
+
+	if a.ContainsAny(71) {
+		t.Error("ContainsSet should not contain 71")
+	}
+
+	if a.ContainsAny(71, 10) {
+		t.Error("ContainsSet should not contain 71 or 10")
+	}
+
+	a.Add(13)
+	a.Add(7)
+	a.Add(1)
+
+	if !(a.ContainsAny(13, 17, 10)) {
+		t.Error("ContainsSet should contain 13, 17, or 10")
+	}
+}
+
 func Test_ClearSet(t *testing.T) {
-	a := makeSet([]int{2, 5, 9, 10})
+	a := makeSetInt([]int{2, 5, 9, 10})
 
 	a.Clear()
 
@@ -224,7 +409,7 @@ func Test_ClearSet(t *testing.T) {
 }
 
 func Test_ClearUnsafeSet(t *testing.T) {
-	a := makeUnsafeSet([]int{2, 5, 9, 10})
+	a := makeUnsafeSetInt([]int{2, 5, 9, 10})
 
 	a.Clear()
 
@@ -234,7 +419,7 @@ func Test_ClearUnsafeSet(t *testing.T) {
 }
 
 func Test_CardinalitySet(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 
 	if a.Cardinality() != 0 {
 		t.Error("set should be an empty set")
@@ -266,7 +451,7 @@ func Test_CardinalitySet(t *testing.T) {
 }
 
 func Test_CardinalityUnsafeSet(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 
 	if a.Cardinality() != 0 {
 		t.Error("set should be an empty set")
@@ -298,9 +483,9 @@ func Test_CardinalityUnsafeSet(t *testing.T) {
 }
 
 func Test_SetIsSubset(t *testing.T) {
-	a := makeSet([]int{1, 2, 3, 5, 7})
+	a := makeSetInt([]int{1, 2, 3, 5, 7})
 
-	b := NewSet()
+	b := NewSet[int]()
 	b.Add(3)
 	b.Add(5)
 	b.Add(7)
@@ -316,10 +501,31 @@ func Test_SetIsSubset(t *testing.T) {
 	}
 }
 
+func Test_SetIsProperSubset(t *testing.T) {
+	a := makeSetInt([]int{1, 2, 3, 5, 7})
+	b := makeSetInt([]int{7, 5, 3, 2, 1})
+
+	if !a.IsSubset(b) {
+		t.Error("set a should be a subset of set b")
+	}
+	if a.IsProperSubset(b) {
+		t.Error("set a should not be a proper subset of set b (they're equal)")
+	}
+
+	b.Add(72)
+
+	if !a.IsSubset(b) {
+		t.Error("set a should be a subset of set b")
+	}
+	if !a.IsProperSubset(b) {
+		t.Error("set a should be a proper subset of set b")
+	}
+}
+
 func Test_UnsafeSetIsSubset(t *testing.T) {
-	a := makeUnsafeSet([]int{1, 2, 3, 5, 7})
+	a := makeUnsafeSetInt([]int{1, 2, 3, 5, 7})
 
-	b := NewThreadUnsafeSet()
+	b := NewThreadUnsafeSet[int]()
 	b.Add(3)
 	b.Add(5)
 	b.Add(7)
@@ -335,15 +541,102 @@ func Test_UnsafeSetIsSubset(t *testing.T) {
 	}
 }
 
-func Test_SetIsSuperSet(t *testing.T) {
-	a := NewSet()
+func Test_UnsafeSetIsProperSubset(t *testing.T) {
+	a := makeUnsafeSetInt([]int{1, 2, 3, 5, 7})
+	b := NewThreadUnsafeSet[int]()
+	b.Add(7)
+	b.Add(1)
+	b.Add(5)
+	b.Add(3)
+	b.Add(2)
+
+	if !a.IsSubset(b) {
+		t.Error("set a should be a subset of set b")
+	}
+	if a.IsProperSubset(b) {
+		t.Error("set a should not be a proper subset of set b (they're equal)")
+	}
+
+	b.Add(72)
+
+	if !a.IsSubset(b) {
+		t.Error("set a should be a subset of set b")
+	}
+	if !a.IsProperSubset(b) {
+		t.Error("set a should be a proper subset of set b because set b has 72")
+	}
+}
+
+func Test_SetIsSuperset(t *testing.T) {
+	a := NewSet[int]()
 	a.Add(9)
 	a.Add(5)
 	a.Add(2)
 	a.Add(1)
 	a.Add(11)
 
-	b := NewSet()
+	b := NewSet[int]()
+	b.Add(5)
+	b.Add(2)
+	b.Add(11)
+
+	if !a.IsSuperset(b) {
+		t.Error("set a should be a superset of set b")
+	}
+
+	b.Add(42)
+
+	if a.IsSuperset(b) {
+		t.Error("set a should not be a superset of set b because set b has a 42")
+	}
+}
+
+func Test_SetIsProperSuperset(t *testing.T) {
+	a := NewSet[int]()
+	a.Add(5)
+	a.Add(2)
+	a.Add(11)
+
+	b := NewSet[int]()
+	b.Add(2)
+	b.Add(5)
+	b.Add(11)
+
+	if !a.IsSuperset(b) {
+		t.Error("set a should be a superset of set b")
+	}
+	if a.IsProperSuperset(b) {
+		t.Error("set a should not be a proper superset of set b (they're equal)")
+	}
+
+	a.Add(9)
+
+	if !a.IsSuperset(b) {
+		t.Error("set a should be a superset of set b")
+	}
+	if !a.IsProperSuperset(b) {
+		t.Error("set a not be a proper superset of set b because set a has a 9")
+	}
+
+	b.Add(42)
+
+	if a.IsSuperset(b) {
+		t.Error("set a should not be a superset of set b because set b has a 42")
+	}
+	if a.IsProperSuperset(b) {
+		t.Error("set a should not be a proper superset of set b because set b has a 42")
+	}
+}
+
+func Test_UnsafeSetIsSuperset(t *testing.T) {
+	a := NewThreadUnsafeSet[int]()
+	a.Add(9)
+	a.Add(5)
+	a.Add(2)
+	a.Add(1)
+	a.Add(11)
+
+	b := NewThreadUnsafeSet[int]()
 	b.Add(5)
 	b.Add(2)
 	b.Add(11)
@@ -359,34 +652,47 @@ func Test_SetIsSuperSet(t *testing.T) {
 	}
 }
 
-func Test_UnsafeSetIsSuperSet(t *testing.T) {
-	a := NewThreadUnsafeSet()
-	a.Add(9)
+func Test_UnsafeSetIsProperSuperset(t *testing.T) {
+	a := NewThreadUnsafeSet[int]()
 	a.Add(5)
 	a.Add(2)
-	a.Add(1)
 	a.Add(11)
 
-	b := NewThreadUnsafeSet()
-	b.Add(5)
+	b := NewThreadUnsafeSet[int]()
 	b.Add(2)
+	b.Add(5)
 	b.Add(11)
 
 	if !a.IsSuperset(b) {
 		t.Error("set a should be a superset of set b")
 	}
+	if a.IsProperSuperset(b) {
+		t.Error("set a should not be a proper superset of set b (they're equal)")
+	}
+
+	a.Add(9)
+
+	if !a.IsSuperset(b) {
+		t.Error("set a should be a superset of set b")
+	}
+	if !a.IsProperSuperset(b) {
+		t.Error("set a not be a proper superset of set b because set a has a 9")
+	}
 
 	b.Add(42)
 
 	if a.IsSuperset(b) {
-		t.Error("set a should not be a superset of set b because set a has a 42")
+		t.Error("set a should not be a superset of set b because set b has a 42")
+	}
+	if a.IsProperSuperset(b) {
+		t.Error("set a should not be a proper superset of set b because set b has a 42")
 	}
 }
 
 func Test_SetUnion(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 
-	b := NewSet()
+	b := NewSet[int]()
 	b.Add(1)
 	b.Add(2)
 	b.Add(3)
@@ -399,30 +705,30 @@ func Test_SetUnion(t *testing.T) {
 		t.Error("set c is unioned with an empty set and therefore should have 5 elements in it")
 	}
 
-	d := NewSet()
+	d := NewSet[int]()
 	d.Add(10)
 	d.Add(14)
 	d.Add(0)
 
 	e := c.Union(d)
 	if e.Cardinality() != 8 {
-		t.Error("set e should should have 8 elements in it after being unioned with set c to d")
+		t.Error("set e should have 8 elements in it after being unioned with set c to d")
 	}
 
-	f := NewSet()
+	f := NewSet[int]()
 	f.Add(14)
 	f.Add(3)
 
 	g := f.Union(e)
 	if g.Cardinality() != 8 {
-		t.Error("set g should still ahve 8 elements in it after being unioned with set f that has duplicates")
+		t.Error("set g should still have 8 elements in it after being unioned with set f that has duplicates")
 	}
 }
 
 func Test_UnsafeSetUnion(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 
-	b := NewThreadUnsafeSet()
+	b := NewThreadUnsafeSet[int]()
 	b.Add(1)
 	b.Add(2)
 	b.Add(3)
@@ -435,33 +741,33 @@ func Test_UnsafeSetUnion(t *testing.T) {
 		t.Error("set c is unioned with an empty set and therefore should have 5 elements in it")
 	}
 
-	d := NewThreadUnsafeSet()
+	d := NewThreadUnsafeSet[int]()
 	d.Add(10)
 	d.Add(14)
 	d.Add(0)
 
 	e := c.Union(d)
 	if e.Cardinality() != 8 {
-		t.Error("set e should should have 8 elements in it after being unioned with set c to d")
+		t.Error("set e should have 8 elements in it after being unioned with set c to d")
 	}
 
-	f := NewThreadUnsafeSet()
+	f := NewThreadUnsafeSet[int]()
 	f.Add(14)
 	f.Add(3)
 
 	g := f.Union(e)
 	if g.Cardinality() != 8 {
-		t.Error("set g should still ahve 8 elements in it after being unioned with set f that has duplicates")
+		t.Error("set g should still have 8 elements in it after being unioned with set f that has duplicates")
 	}
 }
 
 func Test_SetIntersect(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 	a.Add(1)
 	a.Add(3)
 	a.Add(5)
 
-	b := NewSet()
+	b := NewSet[int]()
 	a.Add(2)
 	a.Add(4)
 	a.Add(6)
@@ -483,12 +789,12 @@ func Test_SetIntersect(t *testing.T) {
 }
 
 func Test_UnsafeSetIntersect(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 	a.Add(1)
 	a.Add(3)
 	a.Add(5)
 
-	b := NewThreadUnsafeSet()
+	b := NewThreadUnsafeSet[int]()
 	a.Add(2)
 	a.Add(4)
 	a.Add(6)
@@ -510,12 +816,12 @@ func Test_UnsafeSetIntersect(t *testing.T) {
 }
 
 func Test_SetDifference(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 	a.Add(1)
 	a.Add(2)
 	a.Add(3)
 
-	b := NewSet()
+	b := NewSet[int]()
 	b.Add(1)
 	b.Add(3)
 	b.Add(4)
@@ -531,12 +837,12 @@ func Test_SetDifference(t *testing.T) {
 }
 
 func Test_UnsafeSetDifference(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 	a.Add(1)
 	a.Add(2)
 	a.Add(3)
 
-	b := NewThreadUnsafeSet()
+	b := NewThreadUnsafeSet[int]()
 	b.Add(1)
 	b.Add(3)
 	b.Add(4)
@@ -552,13 +858,13 @@ func Test_UnsafeSetDifference(t *testing.T) {
 }
 
 func Test_SetSymmetricDifference(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 	a.Add(1)
 	a.Add(2)
 	a.Add(3)
 	a.Add(45)
 
-	b := NewSet()
+	b := NewSet[int]()
 	b.Add(1)
 	b.Add(3)
 	b.Add(4)
@@ -574,13 +880,13 @@ func Test_SetSymmetricDifference(t *testing.T) {
 }
 
 func Test_UnsafeSetSymmetricDifference(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 	a.Add(1)
 	a.Add(2)
 	a.Add(3)
 	a.Add(45)
 
-	b := NewThreadUnsafeSet()
+	b := NewThreadUnsafeSet[int]()
 	b.Add(1)
 	b.Add(3)
 	b.Add(4)
@@ -596,8 +902,8 @@ func Test_UnsafeSetSymmetricDifference(t *testing.T) {
 }
 
 func Test_SetEqual(t *testing.T) {
-	a := NewSet()
-	b := NewSet()
+	a := NewSet[int]()
+	b := NewSet[int]()
 
 	if !a.Equal(b) {
 		t.Error("Both a and b are empty sets, and should be equal")
@@ -633,8 +939,8 @@ func Test_SetEqual(t *testing.T) {
 }
 
 func Test_UnsafeSetEqual(t *testing.T) {
-	a := NewThreadUnsafeSet()
-	b := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
+	b := NewThreadUnsafeSet[int]()
 
 	if !a.Equal(b) {
 		t.Error("Both a and b are empty sets, and should be equal")
@@ -670,7 +976,7 @@ func Test_UnsafeSetEqual(t *testing.T) {
 }
 
 func Test_SetClone(t *testing.T) {
-	a := NewSet()
+	a := NewSet[int]()
 	a.Add(1)
 	a.Add(2)
 
@@ -694,7 +1000,7 @@ func Test_SetClone(t *testing.T) {
 }
 
 func Test_UnsafeSetClone(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[int]()
 	a.Add(1)
 	a.Add(2)
 
@@ -717,71 +1023,211 @@ func Test_UnsafeSetClone(t *testing.T) {
 	}
 }
 
-func Test_Iterator(t *testing.T) {
-	a := NewSet()
+func Test_Each(t *testing.T) {
+	a := NewSet[string]()
 
 	a.Add("Z")
 	a.Add("Y")
 	a.Add("X")
 	a.Add("W")
 
-	b := NewSet()
+	b := NewSet[string]()
+	a.Each(func(elem string) bool {
+		b.Add(elem)
+		return false
+	})
+
+	if !a.Equal(b) {
+		t.Error("The sets are not equal after iterating (Each) through the first set")
+	}
+
+	var count int
+	a.Each(func(elem string) bool {
+		if count == 2 {
+			return true
+		}
+		count++
+		return false
+	})
+	if count != 2 {
+		t.Error("Iteration should stop on the way")
+	}
+}
+
+func Test_Iter(t *testing.T) {
+	a := NewSet[string]()
+
+	a.Add("Z")
+	a.Add("Y")
+	a.Add("X")
+	a.Add("W")
+
+	b := NewSet[string]()
 	for val := range a.Iter() {
 		b.Add(val)
 	}
 
 	if !a.Equal(b) {
-		t.Error("The sets are not equal after iterating through the first set")
+		t.Error("The sets are not equal after iterating (Iter) through the first set")
+	}
+}
+
+func Test_UnsafeIter(t *testing.T) {
+	a := NewThreadUnsafeSet[string]()
+
+	a.Add("Z")
+	a.Add("Y")
+	a.Add("X")
+	a.Add("W")
+
+	b := NewThreadUnsafeSet[string]()
+	for val := range a.Iter() {
+		b.Add(val)
+	}
+
+	if !a.Equal(b) {
+		t.Error("The sets are not equal after iterating (Iter) through the first set")
+	}
+}
+
+func Test_Iterator(t *testing.T) {
+	a := NewSet[string]()
+
+	a.Add("Z")
+	a.Add("Y")
+	a.Add("X")
+	a.Add("W")
+
+	b := NewSet[string]()
+	for val := range a.Iterator().C {
+		b.Add(val)
+	}
+
+	if !a.Equal(b) {
+		t.Error("The sets are not equal after iterating (Iterator) through the first set")
 	}
 }
 
 func Test_UnsafeIterator(t *testing.T) {
-	a := NewThreadUnsafeSet()
+	a := NewThreadUnsafeSet[string]()
 
 	a.Add("Z")
 	a.Add("Y")
 	a.Add("X")
 	a.Add("W")
 
-	b := NewThreadUnsafeSet()
-	for val := range a.Iter() {
+	b := NewThreadUnsafeSet[string]()
+	for val := range a.Iterator().C {
 		b.Add(val)
 	}
 
 	if !a.Equal(b) {
-		t.Error("The sets are not equal after iterating through the first set")
+		t.Error("The sets are not equal after iterating (Iterator) through the first set")
 	}
 }
 
-func Test_PowerSet(t *testing.T) {
-	a := NewThreadUnsafeSet()
+func Test_IteratorStop(t *testing.T) {
+	a := NewSet[string]()
 
-	a.Add(1)
-	a.Add("delta")
-	a.Add("chi")
-	a.Add(4)
+	a.Add("Z")
+	a.Add("Y")
+	a.Add("X")
+	a.Add("W")
 
-	b := a.PowerSet()
-	if b.Cardinality() != 16 {
-		t.Error("unexpected PowerSet cardinality")
+	it := a.Iterator()
+	it.Stop()
+	for range it.C {
+		t.Error("The iterating (Iterator) did not stop after Stop() has been called")
+	}
+}
+
+func Test_PopSafe(t *testing.T) {
+	a := NewSet[string]()
+
+	a.Add("a")
+	a.Add("b")
+	a.Add("c")
+	a.Add("d")
+
+	aPop := func() (v string) {
+		v, _ = a.Pop()
+		return
+	}
+
+	captureSet := NewSet[string]()
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	finalNil := aPop()
+
+	if captureSet.Cardinality() != 4 {
+		t.Error("unexpected captureSet cardinality; should be 4")
+	}
+
+	if a.Cardinality() != 0 {
+		t.Error("unepxected a cardinality; should be zero")
+	}
+
+	if !captureSet.Contains("c", "a", "d", "b") {
+		t.Error("unexpected result set; should be a,b,c,d (any order is fine")
+	}
+
+	if finalNil != "" {
+		t.Error("when original set is empty, further pops should result in nil")
+	}
+}
+
+func Test_PopUnsafe(t *testing.T) {
+	a := NewThreadUnsafeSet[string]()
+
+	a.Add("a")
+	a.Add("b")
+	a.Add("c")
+	a.Add("d")
+
+	aPop := func() (v string) {
+		v, _ = a.Pop()
+		return
+	}
+
+	captureSet := NewThreadUnsafeSet[string]()
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	captureSet.Add(aPop())
+	finalNil := aPop()
+
+	if captureSet.Cardinality() != 4 {
+		t.Error("unexpected captureSet cardinality; should be 4")
+	}
+
+	if a.Cardinality() != 0 {
+		t.Error("unepxected a cardinality; should be zero")
+	}
+
+	if !captureSet.Contains("c", "a", "d", "b") {
+		t.Error("unexpected result set; should be a,b,c,d (any order is fine")
+	}
+
+	if finalNil != "" {
+		t.Error("when original set is empty, further pops should result in nil")
 	}
 }
 
 func Test_EmptySetProperties(t *testing.T) {
-	empty := NewSet()
+	empty := NewSet[string]()
 
-	a := NewSet()
-	a.Add(1)
+	a := NewSet[string]()
+	a.Add("1")
 	a.Add("foo")
 	a.Add("bar")
 
-	b := NewSet()
+	b := NewSet[string]()
 	b.Add("one")
 	b.Add("two")
-	b.Add(3)
-	b.Add(4)
-
-	c := NewSet()
+	b.Add("3")
+	b.Add("4")
 
 	if !empty.IsSubset(a) || !empty.IsSubset(b) {
 		t.Error("The empty set is supposed to be a subset of all sets")
@@ -795,7 +1241,7 @@ func Test_EmptySetProperties(t *testing.T) {
 		t.Error("The empty set is supposed to be a subset and a superset of itself")
 	}
 
-	c = a.Union(empty)
+	c := a.Union(empty)
 	if !c.Equal(a) {
 		t.Error("The union of any set with the empty set is supposed to be equal to itself")
 	}
@@ -805,57 +1251,13 @@ func Test_EmptySetProperties(t *testing.T) {
 		t.Error("The intesection of any set with the empty set is supposed to be the empty set")
 	}
 
-	c = a.CartesianProduct(empty)
-	if c.Cardinality() != 0 {
-		t.Error("Cartesian product of any set and the empty set must be the empty set")
-	}
-
 	if empty.Cardinality() != 0 {
 		t.Error("Cardinality of the empty set is supposed to be zero")
 	}
-
-	c = empty.PowerSet()
-	if c.Cardinality() != 1 {
-		t.Error("Cardinality of the power set of the empty set is supposed to be one { {} }")
-	}
-}
-
-func Test_CartesianProduct(t *testing.T) {
-	a := NewThreadUnsafeSet()
-	b := NewThreadUnsafeSet()
-	empty := NewThreadUnsafeSet()
-
-	a.Add(1)
-	a.Add(2)
-	a.Add(3)
-
-	b.Add("one")
-	b.Add("two")
-	b.Add("three")
-	b.Add("alpha")
-	b.Add("gamma")
-
-	c := a.CartesianProduct(b)
-	d := b.CartesianProduct(a)
-
-	if c.Cardinality() != d.Cardinality() {
-		t.Error("Cardinality of AxB must be equal to BxA")
-	}
-
-	if c.Cardinality() != (a.Cardinality() * b.Cardinality()) {
-		t.Error("Unexpected cardinality for cartesian product set")
-	}
-
-	c = a.CartesianProduct(empty)
-	d = empty.CartesianProduct(b)
-
-	if c.Cardinality() != 0 || d.Cardinality() != 0 {
-		t.Error("Cartesian product of any set and the emtpy set Ax0 || 0xA must be the empty set")
-	}
 }
 
 func Test_ToSliceUnthreadsafe(t *testing.T) {
-	s := makeUnsafeSet([]int{1, 2, 3})
+	s := makeUnsafeSetInt([]int{1, 2, 3})
 	setAsSlice := s.ToSlice()
 	if len(setAsSlice) != s.Cardinality() {
 		t.Errorf("Set length is incorrect: %v", len(setAsSlice))
@@ -868,6 +1270,82 @@ func Test_ToSliceUnthreadsafe(t *testing.T) {
 	}
 }
 
+func Test_NewSetFromMapKey_Ints(t *testing.T) {
+	m := map[int]int{
+		5: 5,
+		2: 3,
+	}
+
+	s := NewSetFromMapKeys(m)
+
+	if len(m) != s.Cardinality() {
+		t.Errorf("Length of Set is not the same as the map. Expected: %d. Actual: %d", len(m), s.Cardinality())
+	}
+
+	for k := range m {
+		if !s.Contains(k) {
+			t.Errorf("Element %d not found in map: %v", k, m)
+		}
+	}
+}
+
+func Test_NewSetFromMapKey_Strings(t *testing.T) {
+	m := map[int]int{
+		5: 5,
+		2: 3,
+	}
+
+	s := NewSetFromMapKeys(m)
+
+	if len(m) != s.Cardinality() {
+		t.Errorf("Length of Set is not the same as the map. Expected: %d. Actual: %d", len(m), s.Cardinality())
+	}
+
+	for k := range m {
+		if !s.Contains(k) {
+			t.Errorf("Element %q not found in map: %v", k, m)
+		}
+	}
+}
+
+func Test_NewThreadUnsafeSetFromMapKey_Ints(t *testing.T) {
+	m := map[int]int{
+		5: 5,
+		2: 3,
+	}
+
+	s := NewThreadUnsafeSetFromMapKeys(m)
+
+	if len(m) != s.Cardinality() {
+		t.Errorf("Length of Set is not the same as the map. Expected: %d. Actual: %d", len(m), s.Cardinality())
+	}
+
+	for k := range m {
+		if !s.Contains(k) {
+			t.Errorf("Element %d not found in map: %v", k, m)
+		}
+	}
+}
+
+func Test_NewThreadUnsafeSetFromMapKey_Strings(t *testing.T) {
+	m := map[int]int{
+		5: 5,
+		2: 3,
+	}
+
+	s := NewThreadUnsafeSetFromMapKeys(m)
+
+	if len(m) != s.Cardinality() {
+		t.Errorf("Length of Set is not the same as the map. Expected: %d. Actual: %d", len(m), s.Cardinality())
+	}
+
+	for k := range m {
+		if !s.Contains(k) {
+			t.Errorf("Element %q not found in map: %v", k, m)
+		}
+	}
+}
+
 func Test_Example(t *testing.T) {
 	/*
 	   requiredClasses := NewSet()
@@ -877,7 +1355,7 @@ func Test_Example(t *testing.T) {
 	   requiredClasses.Add("Biology")
 
 	   scienceSlice := []interface{}{"Biology", "Chemistry"}
-	   scienceClasses := NewSetFromSlice(scienceSlice)
+	   scienceClasses := NewSet(scienceSlice)
 
 	   electiveClasses := NewSet()
 	   electiveClasses.Add("Welding")
